import copy
import itertools
import sys
from collections import defaultdict

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given
from hypothesis._strategies import composite
from hypothesis.stateful import RuleBasedStateMachine, Bundle, rule
from pytest import fixture, mark

from negmas.apps.scml import ManufacturingProfile, Product, Process, InputOutput, Job, RunningCommandInfo
from negmas.apps.scml.simulators import SlowFactorySimulator, NO_PRODUCTION, FastFactorySimulator, storage_as_array
from negmas.apps.scml.world import Factory

n_lines = 5
n_levels = 4
n_products = 5
n_processes = 3
initial_wallet = 100.0
n_steps = 100
max_storage = sys.maxsize
initial_storage = dict(zip(range(n_products * n_levels), range(10, 10 * (n_products * n_levels), 10)))


@fixture(scope='module')
def products():
    return [Product(id=i + l * n_levels, production_level=l, name=f'{l}_{i}', catalog_price=(i + 1) * (l + 1),
                    expires_in=None)
            for i in range(n_products) for l in range(n_levels)]


@fixture(scope='module')
def processes():
    return [Process(id=i + l * n_levels, production_level=l, name=f'p{l}_{i}'
                    , inputs={InputOutput(product=i + (l - 1) * n_levels, quantity=3, step=0.0)
            , InputOutput(product=i + (l - 1) * n_levels, quantity=2, step=0.0)}
                    , outputs={InputOutput(product=i + l * n_levels, quantity=1, step=1.0)
            , InputOutput(product=i + l * n_levels, quantity=2, step=1.0)}
                    , historical_cost=1 + i + l * n_levels)
            for i in range(n_processes) for l in range(1, n_levels)]


@fixture(scope='module')
def profiles(processes):
    return [ManufacturingProfile(n_steps=i + 1, cost=10 * (i + 1), initial_pause_cost=1, running_pause_cost=2
                                 , resumption_cost=3, cancellation_cost=4, line=l, process=p)
            for i, (p, l) in enumerate(zip(processes, itertools.cycle(range(n_lines))))]


@fixture(scope='module')
def n_profiles(profiles):
    return len(profiles)


@fixture
def empty_factory(profiles):
    return Factory(id='factory', profiles=profiles, initial_wallet=initial_wallet, initial_storage={}
                   , max_storage=max_storage)


@fixture
def factory_with_storage(profiles):
    return Factory(id='factory', profiles=profiles, initial_storage=copy.deepcopy(initial_storage)
                   , initial_wallet=initial_wallet, max_storage=max_storage)


@fixture
def slow_simulator(profiles, products):
    return SlowFactorySimulator(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                                , n_products=len(products), profiles=profiles, max_storage=max_storage)


def test_can_init_factory(empty_factory):
    assert empty_factory._n_lines == n_lines


def test_schedule_and_job(factory_with_storage):
    profile_ind = 0
    factory = factory_with_storage
    t = 10
    line = factory.profiles[profile_ind].line
    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=True)
    factory.schedule(job=job)
    assert len(factory._jobs) == 1
    assert factory._jobs[(t, factory.profiles[profile_ind].line)] == job
    assert factory._wallet == initial_wallet
    assert len(factory._storage) == len(initial_storage)
    assert factory._storage == initial_storage

    profile = factory.profiles[profile_ind]

    # nothing happens before job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.is_empty for info in infos)
        assert all(command.action == 'none' for command in factory._commands)

    # the command is executed in job time
    for _ in range(profile.n_steps):
        infos = factory.step()
        assert not any(info.failed for info in infos)
        assert all(command.action == 'none' for i, command in enumerate(factory._commands) if i != line)
        if _ == 0:
            assert isinstance(infos[line].started, RunningCommandInfo)
            assert infos[line].started.beg == factory._next_step - 1
            assert infos[line].started.end == profile.n_steps + factory._next_step - 1
        if _ == profile.n_steps - 1:
            assert isinstance(infos[line].finished, RunningCommandInfo)
        if 0 < _ < profile.n_steps - 1:
            assert isinstance(infos[line].continuing, RunningCommandInfo)
            assert not infos[line].finished and not infos[line].started

    # nothing happens after job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.no_production for info in infos)
        assert all(command.action == 'none' for command in factory._commands)


@mark.parametrize('wallet', [initial_wallet, 0], ids=['has_money', 'no_money'])
def test_schedule_and_job_with_failures(empty_factory, wallet):
    profile_ind = 0
    factory = empty_factory
    factory._wallet = wallet
    t = 10
    line = factory.profiles[profile_ind].line
    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=True)
    factory.schedule(job=job)
    assert len(factory._jobs) == 1
    assert factory._jobs[(t, factory.profiles[profile_ind].line)] == job
    assert factory._wallet == wallet
    assert len(factory._storage) == 0
    profile = factory.profiles[profile_ind]

    # nothing happens before job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.is_empty for info in infos)
        assert all(command.action == 'none' for command in factory._commands)

    # the command is executed in job time
    infos = factory.step()
    assert not any(info.failed for i, info in enumerate(infos) if i != line)
    assert infos[line].failed
    assert len(infos[line].failure.missing_inputs) > 0
    if wallet < profile.cost:
        assert infos[line].failure.missing_money == profile.cost - wallet

    # nothing happens after job time
    for _ in range(2 * t):
        infos = factory.step()
        assert all(info.no_production for info in infos)
        assert all(command.action == 'none' for command in factory._commands)


def test_slow_factory_simulator_can_be_initialized(slow_simulator):
    assert slow_simulator is not None


@given(st.integers(min_value=0, max_value=n_steps * 2))
def test_slow_factory_simulator_can_be_checked_at_any_time(slow_simulator, t):
    assert slow_simulator.wallet_at(t) == initial_wallet
    assert slow_simulator.total_storage_at(t) == sum(initial_storage.values())


def do_simulator_run(simulator, profiles, t, at, profile_ind, override):
    profile = profiles[profile_ind]
    line = profile.line
    cost = profile.cost
    length = profile.n_steps
    total_inputs = sum(_.quantity for _ in profile.process.inputs) if at >= t else 0
    total_outputs = sum(_.quantity for _ in profile.process.outputs) if at >= t + profile.n_steps else 0

    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=False)
    simulator.schedule(job=job, override=override)
    assert simulator.wallet_at(at) == (initial_wallet if at < t else initial_wallet - cost)
    assert simulator.total_storage_at(at) == sum(initial_storage.values()) - total_inputs + total_outputs
    assert (simulator.line_schedules_at(at)[_] == NO_PRODUCTION for _ in range(n_lines) if _ != line)
    if at < t or at >= t + length:
        assert simulator.line_schedules_at(at)[line] == NO_PRODUCTION

# @todo correct the slow simulator

@mark.parametrize('profile_ind,t,at_,override,simulator_type',
                  [
                      # (0, 10, 'after', True, 'slow'), (0, 10, 'after', False, 'slow'), (0, 10, 'before', True, 'slow'),
                      # (0, 10, 'before', False, 'slow')
                      # , (0, 10, 'just before', True, 'slow'), (0, 10, 'just before', False, 'slow'),
                      # (0, 10, 'just after', True, 'slow')
                      # , (0, 10, 'just after', False, 'slow')
                      # , (0, 10, 'at', True, 'slow'), (0, 10, 'at', False, 'slow'), (0, 10, 'middle', True, 'slow'),
                      # (0, 10, 'middle', False, 'slow')
                      # , (1, 10, 'after', True, 'slow'), (1, 10, 'after', False, 'slow'), (1, 10, 'before', True, 'slow')
                      # , (1, 10, 'before', False, 'slow')
                      # , (1, 10, 'just before', True, 'slow'), (1, 10, 'just before', False, 'slow'),
                      # (1, 10, 'just after', True, 'slow')
                      # , (1, 10, 'just after', False, 'slow')
                      # , (1, 10, 'at', True, 'slow'), (1, 10, 'at', False, 'slow'), (1, 10, 'middle', True, 'slow'),
                      # (1, 10, 'middle', False, 'slow')
                      # , (2, 10, 'after', True, 'slow'), (2, 10, 'after', False, 'slow'), (2, 10, 'before', True, 'slow')
                      # , (2, 10, 'before', False, 'slow')
                      # , (2, 10, 'just before', True, 'slow'), (2, 10, 'just before', False, 'slow'),
                      # (2, 10, 'just after', True, 'slow')
                      # , (2, 10, 'just after', False, 'slow')
                      # , (2, 10, 'at', True, 'slow'), (2, 10, 'at', False, 'slow'), (2, 10, 'middle', True, 'slow'),
                      # (2, 10, 'middle', False, 'slow')
                      # , (3, 10, 'after', True, 'slow'), (3, 10, 'after', False, 'slow'), (3, 10, 'before', True, 'slow')
                      # , (3, 10, 'before', False, 'slow')
                      # , (3, 10, 'just before', True, 'slow'), (3, 10, 'just before', False, 'slow'),
                      # (3, 10, 'just after', True, 'slow')
                      # , (3, 10, 'just after', False, 'slow')
                      # , (3, 10, 'at', True, 'slow'), (3, 10, 'at', False, 'slow'), (3, 10, 'middle', True, 'slow'),
                      # (3, 10, 'middle', False, 'slow'),
                      (0, 0, 'after', True, 'fast'), (0, 0, 'after', False, 'fast'), (0, 0, 'before', True, 'fast'),
                      (0, 0, 'before', False, 'fast')
                      , (0, 0, 'just before', True, 'fast'), (0, 0, 'just before', False, 'fast'),
                      (0, 0, 'just after', True, 'fast')
                      , (0, 0, 'just after', False, 'fast')
                      , (0, 0, 'at', True, 'fast'), (0, 0, 'at', False, 'fast'), (0, 0, 'middle', True, 'fast'),
                      (0, 0, 'middle', False, 'fast')
                      , (1, 0, 'after', True, 'fast'), (1, 0, 'after', False, 'fast'), (1, 0, 'before', True, 'fast')
                      , (1, 0, 'before', False, 'fast')
                      , (1, 0, 'just before', True, 'fast'), (1, 0, 'just before', False, 'fast'),
                      (1, 0, 'just after', True, 'fast')
                      , (1, 0, 'just after', False, 'fast')
                      , (1, 0, 'at', True, 'fast'), (1, 0, 'at', False, 'fast'), (1, 0, 'middle', True, 'fast'),
                      (1, 0, 'middle', False, 'fast')
                      , (2, 0, 'after', True, 'fast'), (2, 0, 'after', False, 'fast'), (2, 0, 'before', True, 'fast')
                      , (2, 0, 'before', False, 'fast')
                      , (2, 0, 'just before', True, 'fast'), (2, 0, 'just before', False, 'fast'),
                      (2, 0, 'just after', True, 'fast')
                      , (2, 0, 'just after', False, 'fast')
                      , (2, 0, 'at', True, 'fast'), (2, 0, 'at', False, 'fast'), (2, 0, 'middle', True, 'fast'),
                      (2, 0, 'middle', False, 'fast')
                      , (3, 0, 'after', True, 'fast'), (3, 0, 'after', False, 'fast'), (3, 0, 'before', True, 'fast')
                      , (3, 0, 'before', False, 'fast')
                      , (3, 0, 'just before', True, 'fast'), (3, 0, 'just before', False, 'fast'),
                      (3, 0, 'just after', True, 'fast')
                      , (3, 0, 'just after', False, 'fast')
                      , (3, 0, 'at', True, 'fast'), (3, 0, 'at', False, 'fast'), (3, 0, 'middle', True, 'fast'),
                      (3, 0, 'middle', False, 'fast')
                      , (0, 10, 'after', True, 'fast'), (0, 10, 'after', False, 'fast'),
                      (0, 10, 'before', True, 'fast'), (0, 10, 'before', False, 'fast')
                      , (0, 10, 'just before', True, 'fast'), (0, 10, 'just before', False, 'fast'),
                      (0, 10, 'just after', True, 'fast')
                      , (0, 10, 'just after', False, 'fast')
                      , (0, 10, 'at', True, 'fast'), (0, 10, 'at', False, 'fast'), (0, 10, 'middle', True, 'fast'),
                      (0, 10, 'middle', False, 'fast')
                      , (1, 10, 'after', True, 'fast'), (1, 10, 'after', False, 'fast'), (1, 10, 'before', True, 'fast')
                      , (1, 10, 'before', False, 'fast')
                      , (1, 10, 'just before', True, 'fast'), (1, 10, 'just before', False, 'fast'),
                      (1, 10, 'just after', True, 'fast')
                      , (1, 10, 'just after', False, 'fast')
                      , (1, 10, 'at', True, 'fast'), (1, 10, 'at', False, 'fast'), (1, 10, 'middle', True, 'fast'),
                      (1, 10, 'middle', False, 'fast')
                      , (2, 10, 'after', True, 'fast'), (2, 10, 'after', False, 'fast'), (2, 10, 'before', True, 'fast')
                      , (2, 10, 'before', False, 'fast')
                      , (2, 10, 'just before', True, 'fast'), (2, 10, 'just before', False, 'fast'),
                      (2, 10, 'just after', True, 'fast')
                      , (2, 10, 'just after', False, 'fast')
                      , (2, 10, 'at', True, 'fast'), (2, 10, 'at', False, 'fast'), (2, 10, 'middle', True, 'fast'),
                      (2, 10, 'middle', False, 'fast')
                      , (3, 10, 'after', True, 'fast'), (3, 10, 'after', False, 'fast'), (3, 10, 'before', True, 'fast')
                      , (3, 10, 'before', False, 'fast')
                      , (3, 10, 'just before', True, 'fast'), (3, 10, 'just before', False, 'fast'),
                      (3, 10, 'just after', True, 'fast')
                      , (3, 10, 'just after', False, 'fast')
                      , (3, 10, 'at', True, 'fast'), (3, 10, 'at', False, 'fast'), (3, 10, 'middle', True, 'fast'),
                      (3, 10, 'middle', False, 'fast')
                      , (0, 0, 'after', True, 'fast'), (0, 0, 'after', False, 'fast'), (0, 0, 'before', True, 'fast'),
                      (0, 0, 'before', False, 'fast')
                      , (0, 0, 'just before', True, 'fast'), (0, 0, 'just before', False, 'fast'),
                      (0, 0, 'just after', True, 'fast')
                      , (0, 0, 'just after', False, 'fast')
                      , (0, 0, 'at', True, 'fast'), (0, 0, 'at', False, 'fast'), (0, 0, 'middle', True, 'fast'),
                      (0, 0, 'middle', False, 'fast')
                      , (1, 0, 'after', True, 'fast'), (1, 0, 'after', False, 'fast'), (1, 0, 'before', True, 'fast')
                      , (1, 0, 'before', False, 'fast')
                      , (1, 0, 'just before', True, 'fast'), (1, 0, 'just before', False, 'fast'),
                      (1, 0, 'just after', True, 'fast')
                      , (1, 0, 'just after', False, 'fast')
                      , (1, 0, 'at', True, 'fast'), (1, 0, 'at', False, 'fast'), (1, 0, 'middle', True, 'fast'),
                      (1, 0, 'middle', False, 'fast')
                      , (2, 0, 'after', True, 'fast'), (2, 0, 'after', False, 'fast'), (2, 0, 'before', True, 'fast')
                      , (2, 0, 'before', False, 'fast')
                      , (2, 0, 'just before', True, 'fast'), (2, 0, 'just before', False, 'fast'),
                      (2, 0, 'just after', True, 'fast')
                      , (2, 0, 'just after', False, 'fast')
                      , (2, 0, 'at', True, 'fast'), (2, 0, 'at', False, 'fast'), (2, 0, 'middle', True, 'fast'),
                      (2, 0, 'middle', False, 'fast')
                      , (3, 0, 'after', True, 'fast'), (3, 0, 'after', False, 'fast'), (3, 0, 'before', True, 'fast')
                      , (3, 0, 'before', False, 'fast')
                      , (3, 0, 'just before', True, 'fast'), (3, 0, 'just before', False, 'fast'),
                      (3, 0, 'just after', True, 'fast')
                      , (3, 0, 'just after', False, 'fast')
                      , (3, 0, 'at', True, 'fast'), (3, 0, 'at', False, 'fast'), (3, 0, 'middle', True, 'fast'),
                      (3, 0, 'middle', False, 'fast')
                  ]
                  )
def test_slow_factory_simulator_with_jobs(products, profiles, profile_ind, t, at_, override, simulator_type):
    simulator_type = SlowFactorySimulator if simulator_type == 'slow' else FastFactorySimulator
    simulator = simulator_type(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                               , n_products=len(products), profiles=profiles, max_storage=max_storage)
    profile = profiles[profile_ind]
    length = profile.n_steps
    if at_ == 'before':
        if t - 5 < 0:
            return
        at = t - 5
    elif at_ == 'after':
        if t + length + 3 > n_steps:
            return
        at = t + length + 3
    elif at_ == 'at':
        at = t
    elif at_ == 'just before':
        if t - 1 < 0:
            return
        at = t - 1
    elif at_ == 'just after':
        at = t + length
    elif at_ == 'middle':
        if length == 1:
            return
        at = (t + length) // 2
    else:
        raise ValueError(f'Unknown option {at_}')
    do_simulator_run(simulator, profiles, t, at, profile_ind, override)


@given(profile_ind=st.integers(min_value=0, max_value=3), t=st.integers(min_value=0, max_value=n_steps - 1)
    , at=st.integers(min_value=0, max_value=n_steps - 1), override=st.booleans()
    , simulator_type=st.sampled_from(('fast', ))) # @todo add slow back
def test_slow_factory_simulator_with_jobs_hypothesis(profile_ind, t, at, override, simulator_type):
    products = [Product(id=i + l * n_levels, production_level=l, name=f'{l}_{i}', catalog_price=(i + 1) * (l + 1),
                        expires_in=None)
                for i in range(n_products) for l in range(n_levels)]
    processes = [Process(id=i + l * n_levels, production_level=l, name=f'p{l}_{i}'
                         , inputs={InputOutput(product=i + (l - 1) * n_levels, quantity=3, step=0.0)
            , InputOutput(product=i + (l - 1) * n_levels, quantity=2, step=0.0)}
                         , outputs={InputOutput(product=i + l * n_levels, quantity=1, step=1.0)
            , InputOutput(product=i + l * n_levels, quantity=2, step=1.0)}
                         , historical_cost=1 + i + l * n_levels)
                 for i in range(n_processes) for l in range(1, n_levels)]

    profiles = [ManufacturingProfile(n_steps=i + 1, cost=10 * (i + 1), initial_pause_cost=1, running_pause_cost=2
                                     , resumption_cost=3, cancellation_cost=4, line=l, process=p)
                for i, (p, l) in enumerate(zip(processes, itertools.cycle(range(n_lines))))]

    simulator_type = SlowFactorySimulator if simulator_type == 'slow' else FastFactorySimulator
    simulator = simulator_type(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                               , n_products=len(products), profiles=profiles, max_storage=max_storage)
    do_simulator_run(simulator, profiles, t, at, profile_ind, override)


@composite
def storage(draw):
    product = draw(st.integers(0, len(products())))
    quantity = draw(st.integers(-4, 4))
    return product, quantity


sample_products = [Product(id=i + l * n_levels, production_level=l, name=f'{l}_{i}', catalog_price=(i + 1) * (l + 1),
                           expires_in=None)
                   for i in range(n_products) for l in range(n_levels)]
sample_processes = [Process(id=i + l * n_levels, production_level=l, name=f'p{l}_{i}'
                            , inputs={InputOutput(product=i + (l - 1) * n_levels, quantity=3, step=0.0)
        , InputOutput(product=i + (l - 1) * n_levels, quantity=2, step=0.0)}
                            , outputs={InputOutput(product=i + l * n_levels, quantity=1, step=1.0)
        , InputOutput(product=i + l * n_levels, quantity=2, step=1.0)}
                            , historical_cost=1 + i + l * n_levels)
                    for i in range(n_processes) for l in range(1, n_levels)]
sample_profiles = [ManufacturingProfile(n_steps=i + 1, cost=10 * (i + 1), initial_pause_cost=1, running_pause_cost=2
                                        , resumption_cost=3, cancellation_cost=4, line=l, process=p)
                   for i, (p, l) in enumerate(zip(sample_processes, itertools.cycle(range(n_lines))))]


class SimulatorsActAsFactory(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        products, profiles = sample_products, sample_profiles
        self.slow_simulator = SlowFactorySimulator(initial_wallet=initial_wallet, initial_storage=initial_storage
                                                   , n_steps=n_steps
                                                   , n_products=len(products), profiles=profiles,
                                                   max_storage=max_storage)
        self.fast_simulator = FastFactorySimulator(initial_wallet=initial_wallet, initial_storage=initial_storage
                                                   , n_steps=n_steps
                                                   , n_products=len(products), profiles=profiles,
                                                   max_storage=max_storage)
        self.factory = Factory(initial_wallet=initial_wallet, initial_storage=initial_storage
                               , profiles=profiles, max_storage=max_storage)
        self.profiles = self.factory.profiles
        self._payments = defaultdict(list)
        self._transports = defaultdict(list)

    profile_indices = Bundle('profile_indices')
    payments = Bundle('payments')
    times = Bundle('times')
    transports = Bundle('transports')

    @rule(target=profile_indices, k=st.integers(len(sample_profiles)))
    def choose_profile(self, k):
        return k

    @rule(targe=payments, payment=st.floats(-30, 30))
    def choose_payment(self, payment):
        return payment

    @rule(targe=times, payment=st.integers(0, n_steps))
    def choose_time(self, t):
        return t

    @rule(targe=transports, storage=storage())
    def choose_transport(self, t):
        return t

    @rule(profile_index=profile_indices, t=times)
    def run_profile(self, profile_index, t):
        job = Job(profile=profile_index, time=t, line=self.profiles[profile_index].line
                  , action='run', contract=None, override=False)
        end = t + self.profiles[profile_index].n_steps
        if end > n_steps - 1:
            return
        self.factory.schedule(job=job, override=False)
        self.slow_simulator.schedule(job=job, override=False)
        self.fast_simulator.schedule(job=job, override=False)

    @rule(payment=payments, t=times)
    def pay(self, payment, t):
        self._payments[t].append(payment)
        self.slow_simulator.pay(payment, t)
        self.fast_simulator.pay(payment, t)

    @rule(trans=transports, t=times)
    def transport(self, trans, t):
        p, q = trans
        self._transports[t].append(trans)
        self.slow_simulator.transport_to(p, q, t)
        self.fast_simulator.transport_to(p, q, t)

    @rule()
    def run_and_test(self):
        for _ in range(n_steps):
            for payment in self._payments[_]:
                self.factory.pay(payment)
            for p, q in self._transports[_]:
                self.factory.transport_to(p, q)
            self.factory.step()
            assert self.slow_simulator.wallet_at(_) == self.fast_simulator.wallet_at(_) == self.factory.wallet
            assert self.slow_simulator.balance_at(_) == self.fast_simulator.balance_at(_) == self.factory.balance
            assert np.all(self.slow_simulator.storage_at(_) == self.fast_simulator.storage_at(_))
            assert np.all(self.slow_simulator.storage_at(_) == storage_as_array(self.factory.storage
                                                                                , n_products=len(sample_products)))
            assert np.all(self.slow_simulator.line_schedules_at(_) == self.factory.line_schedules)
            assert np.all(self.fast_simulator.line_schedules_at(_) == self.factory.line_schedules)


TestSimulators = SimulatorsActAsFactory.TestCase

if __name__ == '__main__':
    pytest.main(args=[__file__])



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['negmas', 'negmas.apps', 'negmas.apps.scml', 'negmas.external', 'negmas.tests']

package_data = \
{'': ['*'],
 'negmas': ['logs/*'],
 'negmas.tests': ['config/*',
                  'data/10issues/*',
                  'data/AMPOvsCity/*',
                  'data/Laptop/*',
                  'data/Laptop1Issue/*',
                  'data/LaptopConv/*',
                  'data/LaptopConv1D/*',
                  'data/fuzzyagent/*',
                  'data/scenarios/anac/y2010/EnglandZimbabwe/*',
                  'data/scenarios/anac/y2010/ItexvsCypress/*',
                  'data/scenarios/anac/y2010/Travel/*',
                  'data/scenarios/other/S-1NIKFRT-1/*',
                  'scml/*']}

install_requires = \
['Click>=6.0',
 'PyYAML==5.1b1',
 'colorlog>=4.0,<5.0',
 'hypothesis>=4.9,<5.0',
 'inflect>=2.1,<3.0',
 'matplotlib>=3.0,<4.0',
 'numpy>=1.16,<2.0',
 'pandas>=0.24.1,<0.25.0',
 'progressbar2>=3.39,<4.0',
 'py4j>=0.10.8,<0.11.0',
 'pytest-cov>=2.6,<3.0',
 'pytest-runner>=4.4,<5.0',
 'scipy>=1.2,<2.0',
 'setuptools>=40.8,<41.0',
 'stringcase>=1.2,<2.0',
 'tabulate>=0.8.3,<0.9.0',
 'typing>=3.6,<4.0',
 'typing_extensions>=3.7,<4.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses'], 'dask': ['distributed']}

entry_points = \
{'console_scripts': ['negmas = scripts.negmas:cli']}

setup_kwargs = {
    'name': 'negmas',
    'version': '0.1.29',
    'description': 'NEGotiations Managed by Agent Simulations',
    'long_description': "NegMAS\n******\n\n\n.. .. image:: https://img.shields.io/pypi/v/negmas/svg\n..         :target: https://pypi.python.org/pypi/negmas\n\n.. .. image:: https://img.shields.io/travis/yasserfarouk/negmas/svg\n..         :target: https://travis-ci.org/yasserfarouk/negmas\n\n.. .. image:: https://readthedocs.org/projects/negmas/badge/?version=latest\n..         :target: https://negmas/readthedocs.io/en/latest/?badge=latest\n..         :alt: Documentation Status\n\n\n\nA python library for managing autonomous negotiation agents in simulation environments. The name ``negmas`` stands for\neither NEGotiation MultiAgent System or NEGotiations Managed by Agent Simulations (your pick).\n\nIntroduction\n============\n\nThis package was designed to help advance the state-of-art in negotiation research by providing an easy-to-use yet\npowerful platform for autonomous negotiation targeting situated simultaneous negotiations.\nIt grew out of the NEC-AIST collaborative laboratory project.\n\nBy *situated* negotiations, we mean those for which utility functions are not pre-ordained by fiat but are a natural\nresult of a simulated business-like process.\n\nBy *simultaneous* negotiations, we mean sessions of dependent negotiations for which the utility value of an agreement\nof one session is affected by what happens in other sessions.\n\nThe documentation is available at: documentation_\n\n.. _documentation: http://www.yasserm.com/negmas/index.html\n\nMain Features\n=============\n\nThis platform was designed with both flexibility and scalability in mind. The key features of the NegMAS package are:\n\n#. The public API is decoupled from internal details allowing for scalable implementations of the same interaction\n   protocols.\n#. Supports agents engaging in multiple concurrent negotiations.\n#. Provides support for inter-negotiation synchronization either through coupled utility functions or through central\n   *control* agents.\n#. The package provides sample negotiators that can be used as templates for more complex negotiators.\n#. The package supports both mediated and unmediated negotiations.\n#. Supports both bilateral and multilateral negotiations.\n#. Novel negotiation protocols and simulated *worlds* can be added to the package as easily as adding novel negotiators.\n#. Allows for non-traditional negotiation scenarios including dynamic entry/exit from the negotiation.\n#. A large variety of built in utility functions.\n#. Utility functions can be active dynamic entities which allows the system to model a much wider range of dynamic ufuns\n   compared with existing packages.\n#. A distributed system with the same interface and industrial-strength implementation is being created allowing agents\n   developed for NegMAS to be seemingly employed in real-world business operations.\n\nTo use negmas in a project\n\n.. code-block:: python\n\n    import negmas\n\nThe package was designed for many uses cases. On one extreme, it can be used by an end user who is interested in running\none of the built-in negotiation protocols. On the other extreme, it can be used to develop novel kinds of negotiation\nagents, negotiation protocols, multi-agent simulations (usually involving situated negotiations), etc.\n\nRunning existing negotiators/negotiation protocols\n==================================================\n\nUsing the package for negotiation can be as simple as the following code snippet:\n\n.. code-block:: python\n\n    from negmas import SAOMechanism, AspirationNegotiator, MappingUtilityFunction\n    session = SAOMechanism(outcomes=10, n_steps=100)\n    negotiators = [AspirationNegotiator(name=f'a{_}') for _ in range(5)]\n    for negotiator in negotiators:\n        session.add(negotiator, ufun=MappingUtilityFunction(lambda x: random.random() * x[0]))\n\n    session.run()\n\nIn this snippet, we created a mechanism session with an outcome-space of *10* discrete outcomes that would run for *10*\nsteps. Five agents with random utility functions are then created and *added* to the session. Finally the session is\n*run* to completion. The agreement (if any) can then be accessed through the *state* member of the session. The library\nprovides several analytic and visualization tools to inspect negotiations. See the first tutorial on\n*Running a Negotiation* for more details.\n\nDeveloping a negotiator\n=======================\n\nDeveloping a novel negotiator slightly more difficult by is still doable in few lines of code:\n\n.. code-block:: python\n\n    from negmas.negotiators import Negotiator\n    class MyAwsomeNegotiator(Negotiator):\n        def __init__(self):\n            # initialize the parents\n            Negotiator.__init__(self)\n            MultiNegotiationsMixin.__init__(self)\n\n        def respond_(self, offer, state):\n            # decide what to do when receiving an offer @ that negotiation\n            pass\n\n        def propose_(self, state):\n            # proposed the required number of proposals (or less) @ that negotiation\n            pass\n\nBy just implementing `respond_()` and `propose_()`. This negotiator is now capable of engaging in alternating offers\nnegotiations. See the documentation of `Negotiator` for a full description of available functionality out of the box.\n\nDeveloping a negotiation protocol\n=================================\n\nDeveloping a novel negotiation protocol is actually even simpler:\n\n.. code-block:: python\n\n    from negmas.mechanisms import Mechanism\n\n    class MyNovelProtocol(Mechanism):\n        def __init__(self):\n            super().__init__()\n\n        def step_(self):\n            # one step of the protocol\n            pass\n\nBy implementing the single `step_()` function, a new protocol is created. New negotiators can be added to the\nnegotiation using `add()` and removed using `remove()`. See the documentation for a full description of\n`Mechanism` available functionality out of the box [Alternatively you can use `Protocol` instead of `Mechanism`].\n\n\nRunning a world simulation\n==========================\n\nThe *raison d'être* for NegMAS is to allow you to develop negotiation agents capable of behaving in realistic\n*business like* simulated environments. These simulations are called *worlds* in NegMAS. Agents interact with each other\nwithin these simulated environments trying their maximize some intrinsic utility function of the agent through several\n*possibly simultaneous* negotiations.\n\nThe `situated` module provides all that you need to create such worlds. An example can be found in the `scml` package.\nThis package implements a supply chain management system in which factory managers compete to maximize their profits in\na market with only negotiations as the means of securing contracts.\n\n\nAcknowledgement\n===============\n\n.. _Genius: http://ii.tudelft.nl/genius\n\nNegMAS tests use scenarios used in ANAC 2010 to ANAC 2018 competitions obtained from the Genius_ Platform. These domains\ncan be found in the tests/data and notebooks/data folders.\n",
    'author': 'Yasser Mohammad',
    'author_email': 'yasserfarouk@gmail.com',
    'url': 'https://github.com/yasserfarouk/negmas',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import functools
import itertools
import math
from abc import ABC
from random import random, randint
from typing import TYPE_CHECKING

from dataclasses import dataclass
from numpy.random import dirichlet

from negmas.events import Notification
from negmas.helpers import ConfigReader, get_class
from negmas.negotiators import NegotiatorProxy
from negmas.outcomes import Issue
from negmas.sao import AspirationNegotiator
from negmas.situated import Contract
from negmas.situated import RenegotiationRequest
from negmas.utilities import normalize, ComplexWeightedUtilityFunction, MappingUtilityFunction
from .common import SCMLAgent, CFP
from .helpers import pos_gauss

if True: #
    from typing import Dict, Any, List, Optional, Union, Tuple
    from .common import Loan

if TYPE_CHECKING:
    from .awi import SCMLAWI

__all__ = [
    'Consumer',
    'ConsumptionProfile',
    'ScheduleDrivenConsumer'
]


@dataclass
class ConsumptionProfile:
    schedule: Union[int, List[int]] = 0
    underconsumption: float = 0.1
    overconsumption: float = 0.0
    dynamicity: float = 0.0
    cv: float = 0.1

    alpha_q: float = 0.5
    alpha_u: float = 1.0

    beta_q: float = 10.0
    beta_u: float = 10.0

    tau_q: float = 2
    tau_u: float = 0.25

    @classmethod
    def random(cls):
        return ConsumptionProfile(schedule=randint(1, 20), overconsumption=2 * random(), underconsumption=2 * random()
                                  , cv=random(), beta_q=99 * random() + 1, beta_u=99 * random() + 1, tau_q=3 * random()
                                  , tau_u=random())

    def schedule_at(self, time: int) -> int:
        if isinstance(self.schedule, int):
            return self.schedule
        else:
            return self.schedule[time % len(self.schedule)]

    def schedule_within(self, time: Union[int, List[int], Tuple[int, int]]) -> int:
        if isinstance(time, int):
            return self.schedule_at(time)
        if isinstance(time, tuple):
            times = list(range(time[0], time[1] + 1))
        else:
            times = time
        if isinstance(self.schedule, int):
            return self.schedule * len(times)
        return sum(self.schedule_at(t) for t in times)

    def set_schedule_at(self, time: int, value: int, n_steps: int) -> None:
        if isinstance(self.schedule, int):
            self.schedule = [self.schedule] * n_steps
        elif len(self.schedule) < n_steps:
            self.schedule = list(itertools.chain(*([self.schedule] * int(math.ceil(n_steps / len(self.schedule))))))
        self.schedule[time % len(self.schedule)] = value


class Consumer(SCMLAgent, ABC):
    """Base class of all consumer classes"""
    pass


class ScheduleDrivenConsumer(Consumer):
    """Consumer class"""

    MAX_UNIT_PRICE = 1e5
    RELATIVE_MAX_PRICE = 2

    def __init__(self, profiles: Dict[int, ConsumptionProfile] = None
                 , negotiator_type='negmas.sao.AspirationNegotiator'
                 , consumption_horizon: Optional[int] = 20
                 , immediate_cfp_update: bool = True
                 , name=None):
        super().__init__(name=name)
        self.negotiator_type = get_class(negotiator_type, scope=globals())
        self.profiles: Dict[int, ConsumptionProfile] = dict()
        self.secured_quantities: Dict[int, int] = dict()
        if profiles is not None:
            self.set_profiles(profiles=profiles)
        self.consumption_horizon = consumption_horizon
        self.immediate_cfp_update = immediate_cfp_update

    def on_new_cfp(self, cfp: 'CFP') -> None:
        pass  # consumers never respond to CFPs

    def init(self):
        super().init()
        if self.consumption_horizon is None:
            self.consumption_horizon = self.awi.n_steps
        self.interesting_products = list(self.profiles.keys())

    def set_profiles(self, profiles: Dict[int, ConsumptionProfile]):
        self.profiles = profiles if profiles is not None else dict()
        self.secured_quantities = dict(zip(profiles.keys(), itertools.repeat(0)))

    def register_product_cfps(self, p: int, t: int, profile: ConsumptionProfile):
        current_schedule = profile.schedule_at(t)
        product = self.products[p]
        awi: SCMLAWI = self.awi
        if current_schedule <= 0:
            awi.bb_remove(section='cfps', query={'publisher': self.id, 'time': t, 'product_index': p})
            return
        max_price = ScheduleDrivenConsumer.RELATIVE_MAX_PRICE * product.catalog_price if product.catalog_price is not None \
            else ScheduleDrivenConsumer.MAX_UNIT_PRICE
        cfps = awi.bb_query(section='cfps', query={'publisher': self.id, 'time': t, 'product': p})
        if cfps is not None and len(cfps) > 0:
            for _, cfp in cfps.items():
                if cfp.max_quantity != current_schedule:
                    cfp = CFP(is_buy=True, publisher=self.id, product=p
                              , time=t, unit_price=(0, max_price), quantity=(1, current_schedule))
                    awi.bb_remove(section='cfps', query={'publisher': self.id, 'time': t
                        , 'product': p})
                    awi.register_cfp(cfp)
                    break
        else:
            cfp = CFP(is_buy=True, publisher=self.id, product=p
                      , time=t, unit_price=(0, max_price), quantity=(1, current_schedule))
            awi.register_cfp(cfp)

    def step(self):
        super().step()
        if self.consumption_horizon is None:
            horizon = self.awi.n_steps
        else:
            horizon = min(self.awi.current_step + self.consumption_horizon + 1, self.awi.n_steps)
        for p, profile in self.profiles.items():
            for t in range(self.awi.current_step, horizon):  # + self.transportation_delay
                self.register_product_cfps(p=p, t=t, profile=profile)

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    @staticmethod
    def _qufun(outcome: Dict[str, Any], tau: float, profile: ConsumptionProfile):
        """The ufun value for quantity"""
        q, t = outcome['quantity'], outcome['time']
        y = profile.schedule_within(t)
        o = profile.overconsumption
        u = profile.underconsumption
        if q == 0 and y != 0:
            return 0.0
        if y == 0:
            result = -o
        elif q > y:
            result = - o * (((q - y) / y) ** tau)
        elif q < y:
            result = - u * (((y - q) / y) ** tau)
        else:
            result = 1.0
        return math.exp(result)

    def on_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[NegotiatorProxy]:
        profile = self.profiles[cfp.product]
        if profile.cv == 0:
            alpha_u, alpha_q = profile.alpha_u, profile.alpha_q
        else:
            alpha_u, alpha_q = tuple(dirichlet((profile.alpha_u, profile.alpha_q), size=1)[0])
        beta_u = pos_gauss(profile.beta_u, profile.cv)
        tau_u = pos_gauss(profile.tau_u, profile.cv)
        tau_q = pos_gauss(profile.tau_q, profile.cv)
        ufun = normalize(ComplexWeightedUtilityFunction(ufuns=[
            MappingUtilityFunction(mapping=lambda x: 1 - x['unit_price'] ** tau_u / beta_u)
            , MappingUtilityFunction(mapping=functools.partial(ScheduleDrivenConsumer._qufun
                                                               , tau=tau_q, profile=profile))]
            , weights=[alpha_u, alpha_q], name=self.name + '_' + partner)
            , outcomes=cfp.outcomes, infeasible_cutoff=-1500)
        if self.negotiator_type == AspirationNegotiator:
            negotiator = self.negotiator_type(assume_normalized=True, name=self.name + '*' + partner
                                              , aspiration_type='boulware')
        else:
            negotiator = self.negotiator_type(name=self.name + '*' + partner)
        negotiator.name = self.name + '_' + partner
        negotiator.utility_function = ufun
        return negotiator

    def set_renegotiation_agenda(self, contract: Contract
                                 , breaches: List[Dict[str, Any]]) -> Optional[RenegotiationRequest]:
        """
        Received by partners in ascending order of their total breach levels in order to set the
        renegotiation agenda when contract execution fails

        Args:
            contract:
            breaches:

        Returns:

        """
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Dict[str, Any]]
                                         , agenda: RenegotiationRequest) -> Optional[NegotiatorProxy]:
        """
        Called to respond to a renegotiation request

        Args:
            agenda:
            contract:
            breaches:

        Returns:

        """
        return None

    def on_renegotiation_request(self, contract: Contract, cfp: "CFP", partner: str) -> bool:
        return False

    def confirm_loan(self, loan: Loan) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""
        return False

    def sign_contract(self, contract: Contract) -> Optional[str]:
        if contract is None:
            return
        cfp: CFP = contract.annotation['cfp']
        agreement = contract.agreement  # type: ignore
        schedule = self.profiles[cfp.product].schedule_at(agreement['time'])
        if schedule - agreement['quantity'] < 0:
            return None
        return super().sign_contract(contract=contract)

    def on_contract_signed(self, contract: Contract):
        super().on_contract_signed(contract)
        if contract is None:
            return
        cfp: CFP = contract.annotation['cfp']
        agreement = contract.agreement  # type: ignore
        self.secured_quantities[cfp.product] += agreement['quantity']
        old_quantity = self.profiles[cfp.product].schedule_at(agreement['time'])
        new_quantity = old_quantity - agreement['quantity']
        t = agreement['time']
        self.profiles[cfp.product].set_schedule_at(time=t
                                                   , value=new_quantity
                                                   , n_steps=self.awi.n_steps)
        if self.immediate_cfp_update and new_quantity != old_quantity:
            self.register_product_cfps(p=cfp.product, t=t, profile=self.profiles[cfp.product])
        for negotiation in self.running_negotiations.values():
            self.notify(negotiation.negotiator, Notification(type='ufun_modified', data=None))



